/*
 *	lpsm.c: lptͳPSѥꥫɤ򥢥ɥ饤
 *	Linux 2.2.xʹ
 */
#ifndef EXPORT_SYMTAB
# define EXPORT_SYMTAB
#endif
#include <linux/config.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
# define MODVERSIONS
#endif
#if defined(MODVERSIONS) & !defined(__GENSYMS__)
# include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/errno.h>

#include <asm/io.h>		/* outp_b */
#include <asm/delay.h>		/* udelay */
#include <asm/uaccess.h>	/* copy_{from,to}_user */

#include <linux/types.h>	/* fs types such as ino_t */
#include <linux/ioport.h>	/* {check,request,release}_region */

#include "psm.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* I/O access wait time */
#define WAIT_BYTE_NORMAL	77	/* usec */
#define WAIT_BYTE_LONG		2753	/* usec */
/* write wait for inter-frame */
#define WAIT_FRAME_WRITE (20 * HZ / 1000)	/* 20msec */


/* lp_base + LP_DATA */
#define PSM_CLK_BIT	0x01
#define PSM_SEL_BIT	0x02	/* negative logic */
#define PSM_CMD_BIT	0x80	/* positive logic */

/* lp_base + LP_STAT */
#define PSM_ACK_BIT	0x40	/* negative, connected to lpt_ack */
#define PSM_DAT_BIT	0x80	/* inverted, connected to lpt_busy */

#define INLINE	__inline__


/*----------------------------------------------------------------------
 * insmod / modprobe interface
 */

/*	0   No message
 *	1   greeting and error message
 *	2   operation log
 *	3   dumps communication log */
static int verbose = 1;
/* base address of PC-style lpt port */
static int io = LPT_PORT;
/* time of last write operation */
static unsigned long last_write_time;


MODULE_PARM(io, "i");
MODULE_PARM(verbose, "i");


/*----------------------------------------------------------------------
 * local variables
 */
/* base address of PC-style lpt port */
static int linpsm_base;
/* major device number */
static int linpsm_major;
/* lpt data port */
static unsigned char lp_data;


/*======================================================================
 * إѡ롼
 *======================================================================*/

/*
 * I/Oݡȥ
 */
static INLINE void linpsm_write_data(unsigned char data)
{
	lp_data = data;
	outb_p(data, linpsm_base + 0);
}

static INLINE unsigned char linpsm_read_status(void)
{
	return inb_p(linpsm_base + 1);
}

static INLINE void linpsm_write_ctrl(unsigned char data)
{
	outb_p(data, linpsm_base + 2);
}


static int linpsm_attach(void)
{
	linpsm_write_data(0xff);
	linpsm_write_ctrl(0x08);
	udelay(1000);	/* 50000 */
	linpsm_write_ctrl(0x0c);
	last_write_time = jiffies;
	return 0;
}

static void dump_image(int len, const unsigned char *buf)
{
	int i;
	char s[6 + 3 * 16 + 1];

	for(i = 0; i < len; i++) {
		if(i % 16 == 0)
			sprintf(s + 0, "+%03x: ", i);
		sprintf(s + 6 + (i % 16) * 3, "%02x ", buf[i]);
		if(i % 16 == 15)
			printk("%s\n", s);
	}
	if(i % 16 != 0)
		printk("%s\n", s);
}


/*----------------------------------------------------------------------
 * ХȤž롼
 */
static unsigned char linpsm_byteio(unsigned char wdata, int uwait)
{
	int i;
	unsigned char rdata;

	rdata = 0;
	for(i = 0; i < 8; i++) {
		if(wdata & 1)
			lp_data |= PSM_CMD_BIT;
		else
			lp_data &= ~PSM_CMD_BIT;
		wdata >>= 1;
		linpsm_write_data(lp_data & ~PSM_CLK_BIT);
		udelay(5);

		if(linpsm_read_status() & PSM_DAT_BIT)
			rdata = rdata >> 1;
		else
			rdata = (rdata >> 1) | 128;
		linpsm_write_data(lp_data | PSM_CLK_BIT);
		udelay(6);
	}

	/* ̾77usecԤ֤2353usecԤ */
	while(uwait > 1000) {
		udelay(1000);
		uwait -= 1000;
	}
	udelay(uwait);

	return rdata;
}


/*----------------------------------------------------------------------
 * ե졼ɤ߹
 */
static int linpsm_read_frame(int frame, unsigned char buf[PSM_FRAME_SIZE])
{
	int i, t;
	char rcv_buf[PSM_LEN_READ];

	if(verbose > 1) printk(MODULE_NAME ": read frame #%d\n", frame);

	/* /SEL=L CLK=H */
	linpsm_write_data(~PSM_SEL_BIT);
	udelay(350);

	rcv_buf[0] = linpsm_byteio(0x81, WAIT_BYTE_NORMAL);
	rcv_buf[1] = linpsm_byteio('R', WAIT_BYTE_NORMAL);
	rcv_buf[2] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	rcv_buf[3] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	rcv_buf[4] = linpsm_byteio(frame >> 8, WAIT_BYTE_NORMAL);
	rcv_buf[5] = linpsm_byteio(frame & 0xff, WAIT_BYTE_NORMAL);
	rcv_buf[6] = linpsm_byteio(0, WAIT_BYTE_LONG);

	rcv_buf[7] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	rcv_buf[8] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	rcv_buf[9] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	for(i = 0; i < PSM_FRAME_SIZE; i++)
		rcv_buf[i + 10] = linpsm_byteio(buf[i], WAIT_BYTE_NORMAL);
	rcv_buf[138] = linpsm_byteio(0, WAIT_BYTE_NORMAL);
	rcv_buf[139] = linpsm_byteio(0, WAIT_BYTE_NORMAL);

	/* ޥɤαɽ */
	if(verbose > 2) {
		dump_image(PSM_LEN_READ, rcv_buf);
	}

	/* /SEL=H, CLK=H */
	udelay(350);
	linpsm_write_data(0xff);

#ifndef CONFIG_SKIP_ERROR
	/* å */
	if(rcv_buf[2] != 0x5a || rcv_buf[3] != 0x5d
	   || rcv_buf[6] != 0x5c || rcv_buf[7] != 0x5d) {
		if(verbose > 0) printk("linpsm_read_frame: hdr sig illegal, frame=%d\n", frame);
		return -EIO;
	}
	if(rcv_buf[139] != 'G') {
		if(verbose > 0) printk("linpsm_read_frame: end sig illegal, frame=%d\n", frame);
		return -EIO;
	}
	t = 0;
	for(i = 8; i < 8 + PSM_LEN_CHKSUM; i++) t ^= rcv_buf[i];
	if(t != 0) {
		if(verbose > 0) printk("linpsm_read_frame: chksum mismatch %02x, frame=%d\n", t, frame);
		return -EIO;
	}
#endif

	memcpy(buf, rcv_buf + 10, PSM_FRAME_SIZE);

	return 0;
}

/*----------------------------------------------------------------------
 * ե졼ν񤭹
 */
static int linpsm_write_frame(int frame, unsigned char buf[PSM_FRAME_SIZE])
{
	int i, xor;
	char rcv_buf[PSM_LEN_WRITE];	/* 138 bytes buffer */
	long ellapse_time;

	/* some wait between write accesses */
	ellapse_time = jiffies - last_write_time;
	if(ellapse_time < WAIT_FRAME_WRITE) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(WAIT_FRAME_WRITE - ellapse_time);
	}

	/* calc check-sum byte */
	xor = (frame >> 8) ^ frame;
	for(i = 0; i < PSM_FRAME_SIZE; i++)
		xor ^= buf[i];
	xor &= 0xff;

	/* /SEL=L CLK=H */
	linpsm_write_data(~PSM_SEL_BIT);
	udelay(350);

	rcv_buf[0] = linpsm_byteio(0x81, WAIT_BYTE_NORMAL);	/* ? */
	rcv_buf[1] = linpsm_byteio('W', WAIT_BYTE_LONG);		/* ? */
	rcv_buf[2] = linpsm_byteio(0, WAIT_BYTE_NORMAL);		/* 0x5a */
	rcv_buf[3] = linpsm_byteio(0, WAIT_BYTE_NORMAL);		/* 0x5d */
	rcv_buf[4] = linpsm_byteio(frame >> 8, WAIT_BYTE_NORMAL);
	rcv_buf[5] = linpsm_byteio(frame & 0xff, WAIT_BYTE_NORMAL);
	for(i = 0; i < PSM_FRAME_SIZE; i++)
		rcv_buf[i + 6] = linpsm_byteio(buf[i], WAIT_BYTE_NORMAL);
	rcv_buf[134] = linpsm_byteio(xor, WAIT_BYTE_NORMAL);	/* XOR csum */
	rcv_buf[135] = linpsm_byteio(0, WAIT_BYTE_NORMAL);	/* 0x5c */
	rcv_buf[136] = linpsm_byteio(0, WAIT_BYTE_NORMAL);	/* 0x5d */
	rcv_buf[137] = linpsm_byteio(0, WAIT_BYTE_NORMAL);	/* 0x47 */

	/* /SEL=H, CLK=H */
	udelay(350);
	linpsm_write_data(0xff);
	last_write_time = jiffies;

#ifndef CONFIG_SKIP_ERROR
	/* å */
	if(rcv_buf[2] != 0x5a || rcv_buf[3] != 0x5d) {
		if(verbose > 0) printk("linpsm_write_frame: hdr sig illegal, frame=%d\n", frame);
		dump_image(sizeof(rcv_buf), rcv_buf);
		return -EIO;
	}
	if(rcv_buf[135] != 0x5c || rcv_buf[136] != 0x5d) {
		if(verbose > 0) printk("linpsm_write_frame: finish status illegal, frame=%d\n", frame);
		dump_image(sizeof(rcv_buf), rcv_buf);
		return -EIO;
	}
	if(rcv_buf[137] != 'G')  {
		if(verbose > 0) printk("linpsm_write_frame: finish flag illegal, frame=%d\n", frame);
		dump_image(sizeof(rcv_buf), rcv_buf);
		return -EIO;
	}
#endif

	return 0;
}


/*======================================================================
 * ͥ륤󥿡ե롼
 *======================================================================*/

static loff_t linpsm_lseek(struct file *filp, loff_t offset, int orig)
{
	switch (orig) {
	case 0:
		filp->f_pos = offset;
		return filp->f_pos;
	case 1:
		filp->f_pos += offset;
		return filp->f_pos;
	default:
		return -EINVAL;
	}
}

static ssize_t linpsm_read(struct file *filp, char *ubuf, size_t size,
			 loff_t *offset)
{
	int err, count;
	unsigned long ul_offset;
	char buf[PSM_FRAME_SIZE];

	count = 0;
	ul_offset = (unsigned long)*offset;

	if(ul_offset % PSM_FRAME_SIZE)
		/* violating frame boundary */
		return -ENXIO;

	while(size >= PSM_FRAME_SIZE) {
		if(ul_offset >= PSM_MEMORY_SIZE)
			/* reached the end of file */
			break;
		if((err = linpsm_read_frame(ul_offset / PSM_FRAME_SIZE, buf)))
			return err;
		if(copy_to_user(ubuf + count, buf, PSM_FRAME_SIZE))
			return -EFAULT;

		size -= PSM_FRAME_SIZE;
		count += PSM_FRAME_SIZE;
		ul_offset += PSM_FRAME_SIZE;
	}
	*offset = ul_offset;
	return count;
}


static ssize_t linpsm_write(struct file *filp, const char *ubuf, size_t size,
			  loff_t *offset)
{
	int err, count;
	unsigned long ul_offset;
	char buf[PSM_FRAME_SIZE];

	count = 0;
	ul_offset = (unsigned long)*offset;

	if(ul_offset % PSM_FRAME_SIZE)
		/* violating frame boundary */
		return -ENXIO;

	while(size >= PSM_FRAME_SIZE) {
		if(ul_offset >= PSM_MEMORY_SIZE)
			/* reached the end of file */
			break;
		if(copy_from_user(buf, ubuf + count, PSM_FRAME_SIZE))
			return -EFAULT;
		if((err = linpsm_write_frame(ul_offset / PSM_FRAME_SIZE, buf)))
			return err;

		size -= PSM_FRAME_SIZE;
		count += PSM_FRAME_SIZE;
		ul_offset += PSM_FRAME_SIZE;
	}
	*offset = ul_offset;
	return count;
}


static int linpsm_ioctl(struct inode *inode, struct file *filp,
		      unsigned int func, unsigned long arg)
{
	return -EINVAL;
}


static int linpsm_open(struct inode *inode, struct file *filp)
{
	int err;

	/* 2ťץϵĤʤ */
	if(MOD_IN_USE)
		return -EBUSY;

	/* I/OݡȤλѵĤ */
	if((err = check_region(linpsm_base, 3)) < 0)
		return err;
	request_region(linpsm_base, 3, MODULE_NAME);

	/* ݡȽ */
	linpsm_attach();
	MOD_INC_USE_COUNT;
	return 0;
}


static int linpsm_release(struct inode *inode, struct file *filp)
{
	release_region(linpsm_base, 3);
	MOD_DEC_USE_COUNT;
	return 0;
}


/*======================================================================
 * ⥸塼
 *======================================================================*/

static struct file_operations fops = {
	linpsm_lseek,
	linpsm_read,
	linpsm_write,
	NULL, /* readdir */
	NULL, /* poll */
	linpsm_ioctl,
	NULL, /* mmap */
	linpsm_open,
	NULL, /* flush */
	linpsm_release,
	NULL, /* fsync */
	NULL, /* fasync */
	NULL, /* check_media_change */
	NULL, /* revalidate */
	NULL, /* lock */
};


int init_module(void)
{
	EXPORT_NO_SYMBOLS;

	/* major deviceֹ */
	linpsm_base = io;
	linpsm_major = register_chrdev(0, MODULE_NAME, &fops);
	if(linpsm_major < 0)
		return -ENODEV;

	/* Ͽɽ */
	if(verbose)
		printk(MODULE_NAME ": PSX memory card driver loaded, "
		       "version " VERSION ", "
		       "port 0x%x, cdev 0x%x, "
	 	       "compiled in kernel " UTS_RELEASE "\n",
		       linpsm_base, linpsm_major);
	return 0;
}

void cleanup_module(void)
{
	if(verbose > 1)
		printk(MODULE_NAME " unloaded\n");
	unregister_chrdev(linpsm_major, MODULE_NAME);
}
